/*
Eteria IRC Client, an RFC 1459 compliant client program written in Java.
Copyright (C) 2000  Javier Kohen <jkohen at tough.com>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

import java.awt.*;
import java.awt.event.*;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

public class ChatTabs extends Canvas implements MouseListener {
    public final static int NORMAL = 0;
    public final static int CURRENT = 1;
    public final static int ALARM = 2;
    public final static int STATES = 3;

    private final static int TAB_HEIGHT = 20;
    private final static int TAB_ORIG_Y[] = { 0, TAB_HEIGHT };
    private final static int H_GAP = 12;
    private final static int V_GAP = 4;
    private final static int H_SEP = 2;
    private final static int SHIFT_INCREMENT = 50;

    private ActionListener actionListener;

    private FontMetrics old_fm;
    private Image image;
    private Graphics img_gfx;

    private int max_ascent;
    private Vector tabs;
    private Hashtable name_to_tab;
    private Tab current_visible;
    /* The horizontal offset of the tabs in the visible area. */
    private int x_off = 0;

    private Image images[];

    ChatTabs() {
	tabs = new Vector();
	name_to_tab = new Hashtable();

	images = new Image[STATES];

	addMouseListener(this);
    }

    public void setImage(int state, Image img) {
	images[state] = img;
    }

    public void add(String name) {
	Tab tab = new Tab(name);

	tabs.addElement(tab);
	name_to_tab.put(name, tab);

	calculateTabs();
	repaint();
    }

    public void remove(String name) {
	Tab tab = (Tab) name_to_tab.remove(name);
	if (null == tab) {
	    return;
	}

	if (tab.equals(current_visible)) {
	    current_visible = null;
	}
	tabs.removeElement(tab);

	calculateTabs();
	repaint();
    }

    public void rename(String old_name, String name) {
	// FIXME: Must use an atomic operator.
	Tab tab = (Tab) name_to_tab.get(old_name);
	tab.setName(name);
	name_to_tab.put(name, tab);
	name_to_tab.remove(old_name);

	calculateTabs();
	repaint();
    }

    public void setAlarm(String name, boolean alarm) {
	Tab tab = (Tab) name_to_tab.get(name);

	int current_index = tab.getImageIndex();
	int new_index = tab.equals(current_visible) ? ChatTabs.CURRENT
	    : alarm ? ChatTabs.ALARM
	    : ChatTabs.NORMAL;

	if (current_index != new_index) {
	    tab.setImageIndex(new_index);
	    repaint();
	}
    }

    public void shiftLeft() {
	if (x_off > 0) {
	    x_off = Math.max(x_off - SHIFT_INCREMENT, 0);
	}

	repaint();
    }

    public void shiftRight() {
	int comp_width = getSize().width;
	// Find the width of the tabs for both rows.
	int tabs_width[] = new int [2];
	for (Enumeration e = tabs.elements(); e.hasMoreElements(); ) {
	    Tab tab = (Tab) e.nextElement();
	    Rectangle tb = tab.getBounds();

	    int index = tb.y == TAB_ORIG_Y[0] ? 0 : 1;
	    tabs_width[index] = tb.x + tb.width;
	}
	// Get the maximum width for the tabs.
	int tabs_max_width = Math.max(tabs_width[0], tabs_width[1]);

	if (x_off < tabs_max_width - comp_width) {
	    x_off = Math.min(x_off + SHIFT_INCREMENT,
			     tabs_max_width - comp_width - 1);
	}

	repaint();
    }

    public void makeVisible(String name) {
	makeVisible((Tab) name_to_tab.get(name));
    }

    protected void makeVisible(Tab new_visible) {
	int comp_width = getSize().width;
	Rectangle tab_bnd = new_visible.getBounds();

	// Find the width of the tabs for the row this Tab belongs to.
	int tabs_width = 0;
	for (Enumeration e = tabs.elements(); e.hasMoreElements(); ) {
	    Tab tab = (Tab) e.nextElement();
	    Rectangle tb = tab.getBounds();

	    if (tab_bnd.y == tb.y) {
		tabs_width = tb.x + tb.width;
	    }
	}

	if (tabs_width <= comp_width) {
	    x_off = 0;
	} else if (tabs_width - tab_bnd.x >= comp_width) {
	    x_off = tab_bnd.x;
	} else {
	    x_off = tabs_width - comp_width;
	}

	if (null != current_visible) {
	    current_visible.setImageIndex(ChatTabs.NORMAL);
	}
	new_visible.setImageIndex(ChatTabs.CURRENT);

	current_visible = new_visible;
	repaint();
    }

    public String getVisible() {
	return current_visible.getName();
    }

    public String getTabAt(Point p) {
	p.translate(x_off, 0);

	for (Enumeration e = tabs.elements(); e.hasMoreElements(); ) {
	    Tab tab = (Tab) e.nextElement();
	    Rectangle tab_bnd = tab.getBounds();

	    if (tab_bnd.contains(p)) {
		return tab.getName();
	    }
	}

	return null;
    }

    /*
     * This method is synchronized so we don't modify the info during the paint process.
     */
    protected synchronized void calculateTabs() {
	Graphics g = getGraphics();
	FontMetrics fm = g.getFontMetrics();
	max_ascent = fm.getMaxAscent();

	int x[] = { 0, 0 };

	for (Enumeration e = tabs.elements(); e.hasMoreElements(); ) {
	    Tab tab = (Tab) e.nextElement();
	    String name = tab.getName();
	    int row = Channel.isChannel(name) ? 1 : 0;

	    // Trim tab's title if the name's too long.
	    if (name.length() > 20) {
		name = name.substring(0, 18).concat("...");
	    }
	    tab.setTitle(name);

	    int text_width = fm.stringWidth(name);
	    int tab_width = text_width + H_GAP;

	    tab.setBounds(x[row], TAB_ORIG_Y[row], tab_width, TAB_HEIGHT);

	    x[row] += tab_width + H_SEP;
	}
    }

    private void createImageBuffer() {
	Dimension size = getSize();

	image = createImage(size.width, size.height);
	if (null != img_gfx) {
	    img_gfx.dispose();
	}
	img_gfx = image.getGraphics();
    }

    public void update(Graphics g) {
	paint(g);
    }

    /*
     * This method is synchronized so we don't have the info modified in the middle of a paint process.
     */
    public synchronized void paint(Graphics g) {
	FontMetrics fm = g.getFontMetrics();
	if (!fm.equals(old_fm)) {
	    calculateTabs();
	}
	old_fm = fm;

	Dimension size = getSize();

	if (null == img_gfx) {
	    createImageBuffer();
	}

	img_gfx.setColor(getBackground());
	img_gfx.fillRect(0, 0, size.width, size.height);
	img_gfx.setColor(getForeground());

	for (Enumeration e = tabs.elements(); e.hasMoreElements(); ) {
	    Tab tab = (Tab) e.nextElement();
	    String title = tab.getTitle();

	    Image img = images[tab.getImageIndex()];
	    Rectangle tab_bnd = tab.getBounds();
	    int x = tab_bnd.x - x_off;

    	    img_gfx.drawImage(img, x, tab_bnd.y,
			      tab_bnd.width, tab_bnd.height, this);

	    // Render the tab's title.
  	    img_gfx.drawString(title, x + 6, tab_bnd.y + max_ascent + V_GAP);
	}

	g.drawImage(image, 0, 0, this);
    }

    public Dimension getMinimumSize() {
	return getPreferredSize();
    }

    public Dimension getPreferredSize() {
	return new Dimension(320, TAB_HEIGHT * 2);
    }

    public void setBounds(int x, int y, int w, int h) {
	super.setBounds(x, y, w, h);

	if (null != image
	    && (h > image.getHeight(this)
		|| w > image.getWidth(this))) {
	    createImageBuffer();
	}
	makeVisible(current_visible);
    }

    public void addActionListener(ActionListener l) {
	if (null != l) {
	    actionListener = AWTEventMulticaster.add(actionListener, l);
	}
    }

    public void removeActionListener(ActionListener l) {
	if (null != l) {
	    actionListener = AWTEventMulticaster.remove(actionListener, l);
	}
    }

    protected void processActionEvent(ActionEvent e) {
	if (actionListener != null) {
	    actionListener.actionPerformed(e);
	}
    }

    public void mousePressed(MouseEvent ev) {
	if (0 != (ev.getModifiers() & MouseEvent.BUTTON1_MASK) ||
		  0 == ev.getModifiers()) { // WORKAROUND: #32 Netscape Navigator doesn't set it right for BUTTON1.
	    Point p = ev.getPoint();

	    String tab = getTabAt(p);
	    if (null != tab) {
		processActionEvent(new ActionEvent(this,
						   ActionEvent.ACTION_PERFORMED,
						   tab));
	    }
	}
    }

    public void mouseReleased(MouseEvent e) {
    }

    public void mouseClicked(MouseEvent e) {
    }

    public void mouseEntered(MouseEvent e) {
    }

    public void mouseExited(MouseEvent e) {
    }
}

class Tab {
    private String name;
    private String title;
    private int image_index;
    private Rectangle bounds;

    Tab(String name) {
	this.name = name;
	this.image_index = ChatTabs.NORMAL;
    }

    public String getName() {
	return name;
    }

    public void setName(String name) {
	this.name = name;
    }

    public String getTitle() {
	return title;
    }

    public void setTitle(String title) {
	this.title = title;
    }

    public int getImageIndex() {
	return image_index;
    }

    public void setImageIndex(int image_index) {
	this.image_index = image_index;
    }

    public Rectangle getBounds() {
	return bounds;
    }

    public void setBounds(Rectangle r) {
	this.bounds = new Rectangle(r);
    }

    public void setBounds(int x, int y, int w, int h) {
	this.bounds = new Rectangle(x, y, w, h);
    }
}
