/*
Eteria IRC Client, an RFC 1459 compliant client program written in Java.
Copyright (C) 2000  Javier Kohen <jkohen at tough.com>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

import java.text.ParseException;
import java.util.*;

public class Message {
    String prefix;
    String command;
    Vector parameters;

    boolean filterMIRCAttribs = false;

    Message(String msg) throws ParseException {
	StringTokenizer st = new StringTokenizer(msg, " ");
	String t;

	if (!st.hasMoreTokens()) {
	    throw (new ParseException("Empty message", 0));
	}
	t = st.nextToken();

	// Let's see if there's a prefix
	if (t.charAt(0) == ':') {
	    if (t.length() == 1) {
		throw (new ParseException("Empty prefix", 1));
	    }
	    prefix = t.substring(1);
	    if (!st.hasMoreTokens()) {
		throw (new ParseException("Command expected after prefix", 0));
	    }
	    t = st.nextToken();
	} else {
	    prefix = "";
	}

	// Then we should have received the command
	command = t;
	if (!st.hasMoreTokens()) {
	    throw (new ParseException("Parameters expected after command", 0));
	}

	parameters = new Vector();
	st = new StringTokenizer(st.nextToken(""), " ");
	while (st.hasMoreTokens()) {
	    t = st.nextToken();
	    if (t.charAt(0) == ':') {
		// We found the last parameter, it needs special treatment.
		break;
	    }
	    parameters.addElement(t);
	}

	// We want to pass the last parameter unparsed, therefore, we can't use the output of StringTokenizer; we have to work with the original message.
	// Start looking one char past the beginning to avoid finding the prefix marker.
	int colon_index = msg.indexOf(':', 1);
	if (-1 != colon_index) {
	    parameters.addElement(msg.substring(colon_index + 1));
	}
    }

    Message(String command, String parameters[]) {
	this("", command, parameters);
    }

    Message(String prefix, String command, String parameters[]) {
	this.prefix = prefix;
	this.command = command;
	this.parameters = new Vector();
	for (int i = 0; i < parameters.length; i++) {
	    this.parameters.addElement(parameters[i]);
	}
    }

    public String toString() {
	String t = "";

	if (0 != prefix.length()) {
	    t = t.concat(":".concat(prefix).concat(" "));
	}
	t = t.concat(command);
	for (Enumeration e = parameters.elements(); e.hasMoreElements(); ) {
	    t = t.concat(" ");
	    // Force fetch here to see if this is the last parameter
	    String t2 = (String) e.nextElement();
	    if (!e.hasMoreElements()) {
		t = t.concat(":");
	    }
	    t = t.concat(t2);
	}

	return (t.concat("\r\n"));
    }

    // Produce as many 512-byte strings as needed to represent the message, repeating PREFIX and COMMAND as needed
    public Vector toStrings() {
	Vector v = new Vector();

	String base = "";
	String t = "";

	if (0 != prefix.length()) {
	    base = base.concat(":".concat(prefix).concat(" "));
	}
	base = base.concat(command);

	for (Enumeration e = parameters.elements(); e.hasMoreElements(); ) {
	    t = t.concat(" ");
	    // Force fetch here to see if this is the last parameter
	    String t2 = (String) e.nextElement();
	    if (!e.hasMoreElements()) {
		base = base.concat(t).concat(":");
		t = "";
	    }
	    t = t.concat(t2);
	}

	int base_length = base.length();
	int length;
	int i = 0;
	// We have to do this at least once, just for the case of empty parameters
	do {
	    // We use 510 instead of 512 because we add "\r\n" in the loop
	    length = Math.min(510 - base_length, t.length() - i);
	    String slice = t.substring(i, i + length);
	    v.addElement(base.concat(slice).concat("\r\n"));
	    i += length;
	} while (i < t.length());

	return (v);
    }

    public String getPrefix() {
	return (prefix);
    }

    public String getCommand() {
	return (command);
    }

    public String [] getParameters() {
	String a[] = new String [parameters.size()];
	parameters.copyInto(a);

	if (filterMIRCAttribs) {
	    for (int i = 0; i < a.length; i++) {
		a[i] = filterMIRCAttribs(a[i]);
	    }
	}

	return (a);
    }

    public void setFilterMIRCAttribs(boolean state) {
	filterMIRCAttribs = state;
    }

    public boolean getFilterMIRCAttribs() {
	return filterMIRCAttribs;
    }

    /*
      Blamelessly cut-and-pasted from XChat, then ported to Java.
     */
    public static String filterMIRCAttribs(String text) {
        int nc = 0;
        int i = 0, j = 0;
	boolean color = false;
	int len = text.length();
	char new_str[] = new char [len];

	while (len > 0) {
	    char ch = text.charAt(i);

	    if ((color && Character.isDigit(ch) && nc < 2) ||
		(color && ch == ',' && nc < 3)) {
		nc++;
		if (ch == ',') {
		    nc = 0;
		}
	    } else {
		if (color) {
		    color = false;
		}
		switch (ch) {
		case '\003': // ATTR_COLOR
		    color = true;
		    nc = 0;
		    break;
		case '\007': // ATTR_BEEP
		case '\017': // ATTR_RESET
		case '\026': // ATTR_REVERSE
		case '\002': // ATTR_BOLD
		case '\037': // ATTR_UNDERLINE
		    break;
		default:
		    new_str[j] = ch;
		    j++;
		}
	    }
	    i++;
	    len--;
	}

	return new String(new_str, 0, j);
    }
}
