/*
 * Copyright (c) 2000 Lyrisoft Solutions, Inc.
 * Used by permission
 */
package com.lyrisoft.awt;

import java.awt.*;
import java.util.Vector;
import java.util.Enumeration;

/**
 * Subclass of TextView that handles hyperlinks.
 */
public class HyperlinkTextView extends TextView {
    private static final int WWW = 0;
    private static final int CHANNEL = 1;
    private static final int EMAIL = 2;
    private static final int LINK_TYPES = 3;

    private HyperlinkReceiver _hyperlinkReceiver;
    private Frame _frame;

    /**
     * @param autoScroll whether or not to always scroll to the bottom
     * @param receiver the object to notify when hyperlink is clicked
     */
    public HyperlinkTextView(boolean autoScroll, HyperlinkReceiver receiver) {
        super(autoScroll);
        _hyperlinkReceiver = receiver;
    }

    /**
     * helper method used to find the beginning index of an email address
     * @param s the string
     * @param idx the index of the '@' sign
     */
    static int findStartOfEmail(String s, int idx) {
        char[] chars = s.toCharArray();
        for (int i=idx; i >= 0; i--) {
            char c = chars[i];
            if ( Character.isLetterOrDigit(c)
                 || ( i>0 && ( c == '-' || c == '.' || c == '_' || c == '+' ) ) )
            {
                continue;
            } else {
                return i+1;
            }
        }
        return 0;
    }

    /**
     * helper method that finds the end of a hyperlink (or email address)
     * @param s the string
     * @param start the url's starting index
     */
    static int findEndOfHref(String s, int start) {
        char[] chars = s.toCharArray();
        int end = s.length();
        for (int i=start; i < end; i++) {
	    char c = chars[i];
            if (Character.isLetterOrDigit(c))
            {
                continue;
            };
            switch (c) {
            case '.':
            case '?':
            case '&':
            case '=':
            case '-':
            case '_':
            case '/':
            case '#':
            case ':':
            case '~':
            case '%':
            case '+':
                continue;
            default:
                return i;
            }
        }
        return end;
    }

    /**
     * This method finds the index of the first E-mail addresses found in a <code>s</code>.
     *
     * It doesn't accept substrings that match the following "reg-exp" as valid (\b is a word boundary): /(\b@|@\b)/.
     *
     * @param s the string to find in.
     * @return the index of the mail address on success, or -1 if no address is found.
     */
    public static int indexOfEmail(String s) {
	int indexAt = 0;

	while (-1 != (indexAt = s.indexOf('@', indexAt))) {
	    int start = findStartOfEmail(s, indexAt - 1);
	    int end = findEndOfHref(s, indexAt + 1);

	    // Match against the "reg-exp".
	    if (indexAt - start > 0 && indexAt + 1 < end) {
		return start;
	    }

	    indexAt++;
	}

	return -1;
    }

    public TextStyle append(String s, TextStyle style) {
        boolean appendHttp = false;
	int index[] = new int [LINK_TYPES];
	int endIndex = 0;

	TextStyle linkStyle = new TextStyle(style);
	linkStyle.setColor(Color.blue);
	HyperlinkTextRun run = null;

        index[WWW] = s.indexOf("http://");
        if (index[WWW] == -1) {
            index[WWW] = s.indexOf("www.");
            appendHttp = true;
        }

	// We only look for the hash mark here because the other forms of channel name prefixes are less commonly used as channel names, and more commonly used for normal dialog.
	index[CHANNEL] = s.indexOf('#');
	index[EMAIL] = indexOfEmail(s);

	// Find the first occurring link.
	int min_index = Integer.MAX_VALUE;
	int type = LINK_TYPES;
	for (int i = 0; i < LINK_TYPES; i++) {
	    if (-1 != index[i] && index[i] < min_index) {
		min_index = index[i];
		type = i;
	    }
	}

	switch (type) {
	case WWW:
            endIndex = findEndOfHref(s, index[WWW]);
            String text = s.substring(index[WWW], endIndex);
	    String url = text;
            if (appendHttp) {
		url = "http://".concat(text);
            }
            run = new HyperlinkTextRun(text, url, linkStyle);
	    break;
	case CHANNEL:
            endIndex = findEndOfHref(s, index[CHANNEL]);
            String chan = s.substring(index[CHANNEL], endIndex);
	    run = new HyperlinkTextRun(chan, chan, linkStyle);
	    break;
	case EMAIL:
	    int indexAt = s.indexOf('@', index[EMAIL] + 1);
	    endIndex = findEndOfHref(s, indexAt + 1);
	    String addr = s.substring(index[EMAIL], endIndex);
	    run = new HyperlinkTextRun(addr, "mailto:" + addr, linkStyle);
	    break;
	default:
	    // If no link was found, append the text, and return.
	    return super.append(s, style);
	}

	TextStyle mod_style = new TextStyle(style);

	if (index[type] > 0) {
	    // Append the text ahead the link.
	    mod_style = super.append(s.substring(0, index[type]), style);
	}

	if (null != run) {
	    // Append the link.
	    super.append(run);
	}

	if (endIndex != s.length()) {
	    // Process the text following the link.
	    mod_style = append(s.substring(endIndex, s.length()), mod_style);
	}

	return mod_style;
    }

    synchronized TextRun getRun(int x, int y) {
        if (_firstRun == -1) {
            return null;   
        }
        
        int len = _runs.size();
        for (int i = _firstRun; i < len; i++) {
            TextRun run = (TextRun)_runs.elementAt(i);
            
            if (run.contains(x, y, _yTranslation)) {
                return run;
            }
        }
        return null;
    }

    public void addNotify() {
        super.addNotify();
        _frame = getFrame();
    }

    Frame getFrame() {
        Container c = getParent();
        do {
            if (c instanceof Frame) {
                return (Frame)c;
            } else {
                c = c.getParent();
            }
        } while (c != null);
        return null;
    }

    void setCursor(int x, int y) {
        TextRun run = getRun(x, y);
        if (run == null) {
            _frame.setCursor(Frame.DEFAULT_CURSOR);
        } else {
            if (run instanceof HyperlinkTextRun) {
                _frame.setCursor(Frame.HAND_CURSOR);
            } else {
                _frame.setCursor(Frame.DEFAULT_CURSOR);
            }
        }
    }

    public boolean handleEvent(Event e) {
        if (e.id == Event.MOUSE_DOWN) {
            TextRun run = getRun(e.x, e.y);
            if (run == null) {
                return true;
            }
            if (run instanceof HyperlinkTextRun) {
                _hyperlinkReceiver.handleHyperlink(((HyperlinkTextRun)run).getHref());
            }
            return true;
        }
        if (e.id == Event.MOUSE_MOVE) {
            setCursor(e.x, e.y);
            return true;
        }
        return super.handleEvent(e);
    }
}
