/*
 * Copyright (c) 2000 Lyrisoft Solutions, Inc.
 * Used by permission
 */
package com.lyrisoft.awt;

import java.awt.*;
import java.util.Vector;
import java.util.Enumeration;

/**
 * An alternative TextView (non-editable) that allows font colors, sytles,
 * and word-wrap.
 */
public class TextView extends Canvas {
    public static Font DEFAULT_FONT = new Font("Dialog", Font.PLAIN, 10);
    private final static Color colors [] = {
	Color.black,
	Color.lightGray,
	new Color(0, 0, 204),
	new Color(0, 204, 0),
	new Color(203, 0, 0),
	new Color(191, 191, 0),
	new Color(191, 0, 191),
	new Color(255, 167, 0),
	Color.yellow,
	Color.green,
	new Color(0, 203, 203),
	Color.cyan,
	Color.blue,
	Color.magenta,
	new Color(153, 153, 153),
	Color.darkGray
    };

    protected Vector _runs;
    protected TextStyle _defaultStyle;

    private Image _image;
    private Graphics _graphics;
    private Dimension _bufferSize = new Dimension(0, 0);
    private int _fullHeight;
    private ScrollView _scrollView;
    protected int _yTranslation = 0;

    private Point _nextDrawPoint = new Point(0, 0);
    private Dimension _scratchDimension = new Dimension(0, 0);

    private boolean _autoScrolling;

    /* index of the first viewable TextRun */
    protected int _firstRun = 0;

    public TextView(boolean autoScrolling) {
        _runs = new Vector();
        setFont(DEFAULT_FONT);
        _autoScrolling = autoScrolling;
        setBounds(0, 0, 1024, 768);
        _defaultStyle = new TextStyle(getFont(), Color.black);
    }

    public TextView() {
        this(false);
    }

    public TextStyle getDefaultStyle() {
        return _defaultStyle;
    }

    public void setScrollView(ScrollView sv) {
        _scrollView = sv;
    }

    public void setFont(String name) {
        Font oldFont = getFont();
        Font newFont = new Font(name, oldFont.getStyle(), oldFont.getSize());
        setFont(newFont);
    }

    public void setFontSize(int size) {
        Font oldFont = getFont();
        Font newFont = new Font(oldFont.getName(), oldFont.getStyle(), size);
        setFont(newFont);
    }

    public void setFont(Font f) {
        super.setFont(f);
        _defaultStyle = new TextStyle(f, getForeground());
    }

    public TextStyle append(String s, TextStyle style) {
	style = new TextStyle(style);

        if (s.length() == 0) {
            return style;
        }
//          TextRun run = new TextRun(s, style);
//          append(run);

	// Process mIRC text attribs.

	// Style state info.
	TextStyle orig_style = new TextStyle(style);
	Font font = style.getFont();
	String font_name = font.getName();
	int font_style = font.getStyle();
	int font_size = font.getSize();

        int nc = 0;
	int inc = 0;
	int last_index = 0;
	boolean color = false;
	int len = s.length();
	int fg = 0;
//  	int bg = 0;

	for (int i = 0; i < len; i++) {
	    char ch = s.charAt(i);

	    if (color && ((Character.isDigit(ch) && nc < 2)
			  || (',' == ch && nc < 3))) {
		switch (nc) {
		case 0:
		    fg = ch - '0';
		    break;
		case 1:
		    fg = (fg * 10 + ch - '0') % 16;
		    break;
		}
		nc++;
		inc++;
		if (',' == ch) {
		    nc = 0;
		}
	    } else {
		if (color) {
		    style.setColor(colors[fg]);
		    color = false;
		}
		switch (ch) {
		case '\003': // ATTR_COLOR
		    append(new TextRun(s.substring(last_index + inc, i),
				       new TextStyle(style)));
		    last_index = i;
		    color = true;
		    nc = 0;
		    inc = 1;
		    break;
		case '\002': // ATTR_BOLD
		    append(new TextRun(s.substring(last_index + inc, i),
				       new TextStyle(style)));
		    last_index = i;
		    inc = 1;
		    font_style ^= Font.BOLD;
		    style.setFont(new Font(font_name, font_style, font_size));
		    break;
		case '\017': // ATTR_RESET
		    append(new TextRun(s.substring(last_index + inc, i),
				       new TextStyle(style)));
		    last_index = i;
		    inc = 1;
		    style = new TextStyle(orig_style);
		    break;
		case '\037': // ATTR_UNDERLINE
		    append(new TextRun(s.substring(last_index + inc, i),
				       new TextStyle(style)));
		    last_index = i;
		    inc = 1;
		    font_style ^= Font.ITALIC; // Not exactly underline.
		    style.setFont(new Font(font_name, font_style, font_size));
		    break;
		case '\007': // ATTR_BEEP
		case '\026': // ATTR_REVERSE
		    append(new TextRun(s.substring(last_index + inc, i),
				       new TextStyle(style)));
		    last_index = i;
		    inc = 1;
		    break;
		}
	    }
	}

	append(new TextRun(s.substring(last_index + inc, len),
			   new TextStyle(style)));

	return style;
    }

    public synchronized void append(TextRun run) {
	if (0 == run.getText().length()) {
	    return;
	}
        _runs.addElement(run);
        prepare(run);
        if (_scrollView != null) {
            _scrollView.boundsChanged();
            if (_autoScrolling) {
                _scrollView.scrollToBottom();
            }
        }
        repaint();
    }

    public void append(String s) {
        append(s, _defaultStyle);
    }

    public void setYTranslation(int y) {
        _yTranslation = y;
    }

    public void update(Graphics g) {
        paint(g);
    }

    public void paint(Graphics g) {
        renderViewable(g);
    }

    public void setVisible(boolean b) {
	if (b) {
	    if (_autoScrolling) {
		_scrollView.scrollToBottom();
	    }
	}
    }

    public void setBounds(int x, int y, int w, int h) {
        boolean reRender = getSize().width != w;
        super.setBounds(x, y, w, h);
        if (!createBuffer()) {
            return;
        }
        if (reRender) {
            prepareAll();
        }
        if (_scrollView != null) {
            _scrollView.boundsChanged();
            _scrollView.scrollToBottom();
        }
        repaint();
    }

    boolean createBuffer() {
        Dimension size = getSize();
        if (size.width <= 0 || size.height <= 0) {
            return false;
        }
        _image = createImage(size.width, size.height);
        if (_image == null) {
            return false;
        }

        // dispose old graphics if we have one
        if (_graphics != null) {
            _graphics.dispose();
        }
        _graphics = _image.getGraphics();
        return true;
    }

    public Dimension getPreferredSize() {
        return new Dimension(getSize().width, _fullHeight);
    }

    public void prepare(TextRun run) {
        int oldY = _nextDrawPoint.y;
        run.prepare(getSize().width, _scratchDimension, _nextDrawPoint);
        _fullHeight = oldY + _scratchDimension.height;
    }

    public synchronized void prepareAll() {
        _nextDrawPoint.x = 0;
        _nextDrawPoint.y = 0;
        int width = getSize().width;
        for (Enumeration e = _runs.elements(); e.hasMoreElements(); ) {
            TextRun run = (TextRun)e.nextElement();
            prepare(run);
        }
    }

    public synchronized void renderViewable(Graphics g) {
        Dimension size = getSize();

        if (_graphics == null) {
            if (!createBuffer()) {
                return;
            }
        }
        _graphics.setColor(getBackground());
        _graphics.fillRect(0, 0, size.width, size.height);


        // Iterate through all the runs of text.
        TextRun run = null;
        _firstRun = -1;
        int i=0;
        for (Enumeration e = _runs.elements(); e.hasMoreElements(); ) {
            run = (TextRun)e.nextElement();

            // get the bounding rectangle of the next run
            Rectangle r = run.getBoundingRect();

            // Is the TextRun inside the viewable range?
            if (r.y -_yTranslation + r.height >= 0) {
                if (_firstRun == -1) {
                    _firstRun = i;
                }
                // Is the TextRun below the screen?
                if (r.y - _yTranslation > size.height) {
                    // Yes, we're done
                    break;
                }

                // Yes, draw it.
                run.draw(_graphics, -_yTranslation);
            }
            i++;
        }
        g.drawImage(_image, 0, 0, this);
    }
}
